from rest_framework import serializers
from django.core.exceptions import ValidationError
from typing import Any


class DjeasyModelSerializer(serializers.ModelSerializer):

    class Meta:
        model = None
        fields = "__all__"

    @classmethod
    def options(
        cls,
        model_name,
        queryset=None,
        fields="__all__",
        many=True,
        related_fields=None,
        *args: Any,
        **kwargs: Any,
    ):
        return cls(
            model=model_name,
            fields=fields,
            instance=queryset,
            many=many,
            related_fields=related_fields,
            *args,
            **kwargs,
        )

    def __init__(
        self,
        model=None,
        fields="__all__",
        related_fields=None,
        *args: Any,
        **kwargs: Any,
    ):
        self.display_fields = kwargs.pop("display_fields", None)
        self.access_fields = kwargs.pop("access_fields", None)
        self.rename_fields = kwargs.pop("rename_fields", None)

        if model and self.Meta.model is None:
            self.Meta.model = model
            self.Meta.fields = fields

        super(DjeasyModelSerializer, self).__init__(*args, **kwargs)
        self.get_display_fields()
        self.get_access_fields()
        self.get_renaming_fields()

        if related_fields:
            for related_field, related_field_options in related_fields.items():
                self.fields[related_field] = self.serializer_related_to_field(
                    **related_field_options
                )

    def get_display_fields(self):
        if self.display_fields is not None:
            allowed = set(self.display_fields)
            existing = set(self.fields)
            for field_name in existing - allowed:
                self.fields.pop(field_name)

    def get_access_fields(self):
        if self.access_fields:
            for field_name, access_type in self.access_fields.items():
                if access_type not in ["read_only", "write_only"]:
                    raise ValidationError(
                        "The type should be either 'read_only' or 'write_only'"
                    )
                if access_type == "read_only" and field_name in self.fields:
                    self.fields[field_name].read_only = True
                elif access_type == "write_only" and field_name in self.fields:
                    self.fields[field_name].write_only = True

    def get_renaming_fields(self):
        if self.rename_fields:
            for old_name, new_name in self.rename_fields.items():
                if old_name in self.fields:
                    self.fields[new_name] = self.fields.pop(old_name)



from rest_framework.response import Response
from rest_framework import status
from enum import StrEnum


class ResponseStatus(StrEnum):
    SUCCESS = "Success"
    FAILURE = "Failure"
    IN_PROGRESS = "In Progress"
    NOT_FOUND = "Not Found"
    CONFLICT = "Conflict"
    UNAUTHORIZED = "Unauthorized"
    FORBIDDEN = "Forbidden"
    INVALID_INPUT = "Invalid Input"
    TIMEOUT = "Timeout"
    UNAVAILABLE = "Service Unavailable"

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


def SuccessResponse(data):
    return Response(
        {"status": ResponseStatus.SUCCESS, "data": data},
        status=status.HTTP_200_OK,
    )


def FailureResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_203_NON_AUTHORITATIVE_INFORMATION,
    )


def CreatedResponse(data):
    return Response(
        {"status": ResponseStatus.SUCCESS, "data": data}, status=status.HTTP_201_CREATED
    )


def BadRequestResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_400_BAD_REQUEST,
    )


def UnauthorizedResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_401_UNAUTHORIZED,
    )


def ForbiddenResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_403_FORBIDDEN,
    )


def NotFoundResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_404_NOT_FOUND,
    )


def ConflictResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_409_CONFLICT,
    )


def ServerErrorResponse(data):
    return Response(
        {"status": ResponseStatus.FAILURE, "data": data},
        status=status.HTTP_500_INTERNAL_SERVER_ERROR,
    )


def InProgressResponse(data):
    return Response(
        {"status": ResponseStatus.IN_PROGRESS, "data": data},
        status=status.HTTP_102_PROCESSING,
    )


def NotFoundResponse(data):
    return Response(
        {"status": ResponseStatus.NOT_FOUND, "data": data},
        status=status.HTTP_404_NOT_FOUND,
    )


def UnauthorizedResponse(data):
    return Response(
        {"status": ResponseStatus.UNAUTHORIZED, "data": data},
        status=status.HTTP_401_UNAUTHORIZED,
    )


def ForbiddenResponse(data):
    return Response(
        {"status": ResponseStatus.FORBIDDEN, "data": data},
        status=status.HTTP_403_FORBIDDEN,
    )


def InvalidInputResponse(data):
    return Response(
        {"status": ResponseStatus.INVALID_INPUT, "data": data},
        status=status.HTTP_422_UNPROCESSABLE_ENTITY,
    )


def TimeoutResponse(data):
    return Response(
        {"status": ResponseStatus.TIMEOUT, "data": data},
        status=status.HTTP_504_GATEWAY_TIMEOUT,
    )


def ServiceUnavailableResponse(data):
    return Response(
        {"status": ResponseStatus.UNAVAILABLE, "data": data},
        status=status.HTTP_503_SERVICE_UNAVAILABLE,
    )



import io

from django.core.paginator import Paginator
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer

from rest_framework.response import Response
from rest_framework.serializers import Serializer
from django.db.models import QuerySet
from .response import SuccessResponse


def paginator(serializer, per_page):
    return Paginator(message_parser(serializer), per_page=per_page)


# ordered dict to json converter
def message_parser(data):
    result_data = JSONRenderer().render(data)
    init_streams = io.BytesIO(result_data)
    return JSONParser().parse(init_streams)


def paginated_views(
    page_size: str, page_number: str, serializer: Serializer, queryset: QuerySet
) -> Response:
    if page_number and page_size:
        paginator_record = paginator(serializer, page_size)
        paginated_data = paginator_record.page(page_number).object_list
        return SuccessResponse(
            {
                "total_count": queryset.count(),
                "queryset_count": len(paginated_data),
                "data": paginated_data,
            }
        )



from setuptools import find_packages, setup


# read from README.RST
def readme():
    with open("readme.md") as f:
        return f.read()


# updated
setup(
    name="djeasyview",
    version="1.0.15",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Anand Raj",
    author_email="anand98.ar@gmail.com",
    python_requires=">=3.8",
    license="MIT",
    packages=find_packages(),
    install_requires=["django", "djangorestframework"],
)



from typing import Dict, List, Optional, Type, Union

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from rest_framework.generics import ListCreateAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer
from .utils import paginated_views


class BaseMixin:
    model: Type[Model]
    context: Optional[Dict[str, any]] = None
    serializer_save: Optional[Dict[str, any]] = None
    select_related: Optional[List[str]] = None
    prefetch_related: Optional[List[str]] = None
    enable_cache: Optional[bool] = False
    cache_duration: Optional[int] = None
    order_by: Optional[List[str]] = None
    query_params: Optional[Dict] = None
    pagination: Optional[bool] = False

    def __init__(self) -> None:
        if self.enable_cache and not self.cache_duration:
            raise ImproperlyConfigured(
                "if enable cache is set to True you must define the cache duration , example 60 * 60 "
            )

    @method_decorator(cache_page(cache_duration))
    @method_decorator(vary_on_cookie)
    def cached_get(
        self, request: Request, serializer, pk=None, *args, **kwargs
    ) -> Response:
        return self.get_response(
            serializer, self.get_queryset(pk) if pk else self.get_queryset()
        )

    @staticmethod
    def get_response(serializer_klass, queryset) -> Response:
        return Response({"status": "Success", "data": serializer_klass(queryset).data})


class DjeasyListCreateAPI(BaseMixin, ListCreateAPIView):
    create_serializer_class: Type[BaseSerializer]
    list_serializer_class: Type[BaseSerializer]

    def get_create_serializer(self, data: Dict[str, any]) -> BaseSerializer:
        return self.create_serializer_class(data=data)

    def get_queryset(self) -> Model:
        queryset = super().get_queryset()
        if self.select_related:
            queryset = queryset.select_related(*self.select_related)
        if self.prefetch_related:
            queryset = queryset.prefetch_related(*self.prefetch_related)
        if self.query_params:
            reverse_base = {v: k for k, v in self.query_params.items()}
            queryset = queryset.filter(
                **{
                    reverse_base[param_key]: param_value[0]
                    for param_key, param_value in dict(
                        self.request.query_params
                    ).items()
                    if param_key in reverse_base
                }
            )

        return queryset.order_by(*self.order_by) if self.order_by else queryset

    def get_list_serializer(self, instance: Model) -> BaseSerializer:
        return (
            self.list_serializer_class(instance, context=self.context, many=True)
            if self.context
            else self.list_serializer_class(instance, many=True)
        )

    def list(self, request: Request, *args, **kwargs) -> Response:
        if self.enable_cache:
            return self.cached_get(request, self.get_list_serializer, *args, **kwargs)
        if self.pagination:
            page_size: str = self.request.query_params.get("page_size", None)
            page_number: str = self.request.query_params.get("page_number", None)
            if page_size and page_number:
                queryset = self.get_queryset()
                return paginated_views(
                    page_size,
                    page_number,
                    self.get_list_serializer(queryset).data,
                    queryset,
                )
        return self.get_response(self.get_list_serializer, self.get_queryset())

    def post(self, request: Request, *args, **kwargs) -> Response:
        serializer = self.get_create_serializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            (
                serializer.save(**self.serializer_save)
                if self.serializer_save
                else serializer.save()
            )
            return Response({"status": "Success", "data": serializer.data})
        return Response({"status": "Failure", "data": serializer.errors})


class DjeasyRetrieveUpdateAPI(BaseMixin, RetrieveUpdateDestroyAPIView):
    retrieve_serializer_class: Type[BaseSerializer]
    update_serializer_class: Type[BaseSerializer]
    cache_duration: Optional[int] = None

    def get_retrieve_serializer(self, instance: Model) -> BaseSerializer:
        return (
            self.retrieve_serializer_class(instance, context=self.context)
            if self.context
            else self.retrieve_serializer_class(instance)
        )

    def get_update_serializer(
        self, instance: Model, data: Dict[str, any]
    ) -> BaseSerializer:
        return self.update_serializer_class(instance, data=data)

    def get_queryset(self, pk: Union[int, str]) -> Model:
        queryset = super().get_queryset()
        if self.select_related:
            queryset = self.model.objects.select_related(*self.select_related)
        if self.prefetch_related:
            queryset = self.model.objects.prefetch_related(*self.prefetch_related)
        else:
            queryset = self.model
        return get_object_or_404(queryset, pk=pk)

    def retrieve(
        self, request: Request, pk: Union[int, str] = None, *args, **kwargs
    ) -> Response:
        if self.enable_cache:
            return self.cached_get(
                request, self.get_retrieve_serializer, pk, *args, **kwargs
            )
        return self.get_response(self.get_retrieve_serializer, self.get_queryset(pk))

    def update(
        self, request: Request, pk: Union[int, str] = None, *args, **kwargs
    ) -> Response:
        instance = self.get_queryset(pk)
        serializer = self.get_update_serializer(instance, request.data)
        if serializer.is_valid():
            (
                serializer.save(**self.serializer_save)
                if self.serializer_save
                else serializer.save()
            )
            return Response({"status": "Success", "data": serializer.data})
        return Response({"status": "Failure", "data": serializer.errors})

    def delete(
        self, request: Request, pk: Union[int, str] = None, *args, **kwargs
    ) -> Response:
        self.model.objects.filter(id=pk).delete()
        return Response({"status": "Success", "data": "Deleted Successfully"})
